#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/deploy.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 05:31
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from time import sleep
from typing import Dict, List

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import APIError, NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich.progress import track

### Local modules ###
from src.configs import CLUSTERS, IMAGES, NETWORK, PERIPHERALS
from src.schemas import MutexOption, NewAddress, Service, ServiceName


@command
@option("--duo", alternatives=["uno"], cls=MutexOption, is_flag=True, type=bool)
@option("--postgres", is_flag=True, help="Deploy postgres peripheral service", type=bool)
@option("--redis", is_flag=True, help="Deploy redis peripheral service", type=bool)
@option("--uno", alternatives=["duo"], cls=MutexOption, is_flag=True, type=bool)
def deploy(duo: bool, uno: bool, postgres: bool, redis: bool) -> None:
    """Deploy cluster, either with one or two LND nodes."""
    duo = duo or (not duo and not uno)  # defaults to duo network
    cluster: Dict[ServiceName, Service] = (CLUSTERS["duo"], CLUSTERS["uno"])[uno]
    peripheral_select: Dict[str, bool] = {"postgres": postgres, "redis": redis}
    peripherals: Dict[ServiceName, Service] = {
        f"aesir-{k}": v[f"aesir-{k}"] for k, v in PERIPHERALS.items() if peripheral_select[k]  # type: ignore[index, misc]
    }
    cluster.update(peripherals)
    client: DockerClient = from_env()
    if client.ping():
        try:
            client.networks.create(NETWORK, check_duplicate=True)
        except APIError:
            pass
        for name, service in track(cluster.items(), "Deploy specified local cluster:".ljust(42)):
            image_name: str = dict(**IMAGES["required"], **IMAGES["optional"])[service.alias]
            ports: Dict[str, str] = dict(
                map(lambda item: (item[0], item[1]), [port.split(":") for port in service.ports])
            )
            client.containers.run(
                image_name,
                command=service.command,
                detach=True,
                environment=service.env_vars,
                name=name,
                network=NETWORK,
                ports=ports,
            )
        sleep(3)  # wait until lnd ready
        mining_targets: List[str] = []
        for container in track(client.containers.list(), "Generate addresses:".ljust(42)):
            if match(r"aesir-lnd|aesir-ping|aesir-pong", container.name) is not None:
                new_address: NewAddress = TypeAdapter(NewAddress).validate_json(
                    container.exec_run(
                        """
                        lncli
                            --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                            --rpcserver=localhost:10001
                            --tlscertpath=/home/lnd/.lnd/tls.cert
                        newaddress p2wkh
                        """
                    ).output
                )
                mining_targets.append(new_address.address)
        bitcoind: Container
        try:
            bitcoind = client.containers.get("aesir-bitcoind")
        except NotFound:
            print('!! Unable to find "aesir-bitcoind" container.')
            return
        for address in track(mining_targets, "Mine initial capital for parties:".ljust(42)):
            bitcoind.exec_run(
                """
                bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir generatetoaddress 101 %s
                """
                % address
            )


__all__ = ["deploy"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/clean.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 05:31
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match

### Third-party packages ###
from click import Context, command, pass_context
from docker import DockerClient, from_env
from docker.errors import NotFound
from docker.models.networks import Network
from rich.progress import track

### Local modules ###
from src.configs import NETWORK


@command
@pass_context
def clean(context: Context) -> None:
    """Remove all active "aesir-*" containers, drop network."""
    client: DockerClient = from_env()
    if client.ping():
        for container in track(client.containers.list(), "Remove active containers:".ljust(42)):
            if match(r"aesir-*", container.name) is not None:
                container.stop()
                container.remove()
        try:
            aesir_network: Network = client.networks.get(NETWORK)
            aesir_network.remove()
        except NotFound:
            pass


__all__ = ["clean"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/setup.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 06:18
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, List, Set

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from rich.progress import track

### Local modules ###
from src.configs import IMAGES


@command
@option("--postgres", is_flag=True, help="Pull postgres optional image", type=bool)
@option("--redis", is_flag=True, help="Pull redis optional image", type=bool)
def setup(postgres: bool, redis: bool) -> None:
    """Download docker images used by command-line interface."""
    client: DockerClient = from_env()
    if client.ping():
        outputs: List[str] = []
        docker_images: Set[str] = {image.tags[0] for image in client.images.list()}
        for registry_id in track(IMAGES["required"].values(), "Pull required images:".ljust(42)):
            if registry_id in docker_images:
                outputs.append(f"<Image: '{ registry_id }'> already exists in local docker images.")
            else:
                repository, tag = registry_id.split(":")
                client.images.pull(repository=repository, tag=tag)
                outputs.append(f"<Image: '{ registry_id }'> downloaded.")
        list(map(print, outputs))
        ### Optionals ###
        optional_select: Dict[str, bool] = {"postgres": postgres, "redis": redis}
        optionals: List[str] = [
            registry for alias, registry in IMAGES["optional"].items() if optional_select[alias]
        ]
        if len(optionals) == 0:
            return
        outputs = []
        for registry_id in track(optionals, "Pull optional images flagged:".ljust(42)):
            if registry_id in docker_images:
                outputs.append(f"<Image: '{ registry_id }'> already exists in local docker images.")
            else:
                repository, tag = registry_id.split(":")
                client.images.pull(repository=repository, tag=tag)
                outputs.append(f"<Image: '{ registry_id }'> downloaded.")
        list(map(print, outputs))
    else:
        print("!! Unable to connect to docker daemon.")


__all__ = ["setup"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/schemas/service.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 05:31
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import List, Literal

### Third-party packages ###
from pydantic import BaseModel, StrictStr

### Local modules ###
from src.schemas.image import ImageAlias


class Service(BaseModel):
    alias: ImageAlias
    command: List[StrictStr] = []
    env_vars: List[StrictStr] = []
    ports: List[StrictStr]


ServiceName = Literal[
    "aesir-bitcoind",
    "aesir-lnd",
    "aesir-ping",
    "aesir-pong",
    "aesir-postgres",
    "aesir-redis",
]

__all__ = ["Service", "ServiceName"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/mine.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 05:31
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import List

### Third-party packages ###
from apscheduler.schedulers.background import BackgroundScheduler
from click import argument, command
from docker import DockerClient, from_env
from docker.errors import NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.progress import track
from rich.table import Table
from rich.text import Text

### Local modules ###
from src.schemas import BlockchainInfo, LNDInfo, NewAddress


@command
@argument("blockcount", default=1, type=int)
@argument("blocktime", default=5, type=int)
def mine(blockcount: int, blocktime: int) -> None:
    """Scheduled mining with "blockcount" and "blocktime"."""
    client: DockerClient = from_env()
    if client.ping():
        mining_targets: List[str] = []
        for container in track(client.containers.list(), "Generate mining treasuries:".ljust(42)):
            if match(r"aesir-lnd|aesir-ping|aesir-pong", container.name) is not None:
                new_address: NewAddress = TypeAdapter(NewAddress).validate_json(
                    container.exec_run(
                        """
                        lncli
                            --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                            --rpcserver=localhost:10001
                            --tlscertpath=/home/lnd/.lnd/tls.cert
                        newaddress p2wkh
                        """
                    ).output
                )
                mining_targets.append(new_address.address)
        bitcoind: Container
        try:
            bitcoind = client.containers.get("aesir-bitcoind")
        except NotFound:
            print('!! Unable to find "aesir-bitcoind" container.')
            return
        scheduler: BackgroundScheduler = BackgroundScheduler()
        for address in mining_targets:
            scheduler.add_job(
                bitcoind.exec_run,
                "interval",
                [
                    """
                    bitcoin-cli
                        -regtest
                        -rpcuser=aesir
                        -rpcpassword=aesir
                    generatetoaddress %d %s
                    """
                    % (blockcount, address)
                ],
                seconds=blocktime,
            )
        scheduler.start()
        pane: Layout = Layout()
        sidebar: Layout = Layout(size=24)
        containers: Layout = Layout(name="containers", size=20)
        main: Layout = Layout(size=72)
        body: Layout = Layout(name="body", minimum_size=4, ratio=8, size=17)
        footer: Layout = Layout(name="footer", size=3)
        pane.split_row(sidebar, main)
        main.split_column(body, footer)
        sidebar.split_column(containers)
        aesir_containers: List[str] = list(
            map(
                lambda container: container.name,
                filter(
                    lambda container: match(r"aesir-*", container.name), client.containers.list()
                ),
            )
        )
        with Live(pane, refresh_per_second=4, transient=True):
            pane["containers"].update(
                Panel(Text("\n".join(aesir_containers)), title="containers")
            )
            while True:
                ### Update ###
                blockchain_info: BlockchainInfo = TypeAdapter(BlockchainInfo).validate_json(
                    bitcoind.exec_run(
                        """
                        bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir getblockchaininfo
                        """
                    ).output
                )
                names: List[str] = []
                lnd_infos: List[LNDInfo] = []
                for container in client.containers.list():
                    if match("aesir-lnd|aesir-ping|aesir-pong", container.name) is not None:
                        names.append(container.name)
                        lnd_info: LNDInfo = TypeAdapter(LNDInfo).validate_json(
                            container.exec_run(
                                """
                                lncli
                                    --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                                    --rpcserver=localhost:10001
                                    --tlscertpath=/home/lnd/.lnd/tls.cert
                                getinfo
                                """
                            ).output
                        )
                        lnd_infos.append(lnd_info)

                ### Draw ###
                body_table: Table = Table(expand=True, show_lines=True)
                body_table.add_column("Name", justify="left", style="green")
                body_table.add_column("Nodekey", justify="left", style="cyan")
                body_table.add_column("Channels")
                body_table.add_column("Peers")
                body_table.add_column("Height")
                body_table.add_column("Synced?", justify="right")
                for i, lnd_info in enumerate(lnd_infos):
                    body_table.add_row(
                        names[i],
                        "\n".join(
                            lnd_info.identity_pubkey[c : c + 11]
                            for c in range(0, len(lnd_info.identity_pubkey), 11)
                        ),
                        f"{lnd_info.num_active_channels}",
                        f"{lnd_info.num_peers}",
                        f"{lnd_info.block_height}",
                        ("[red]false", "[green]true")[lnd_info.synced_to_chain],
                    )
                pane["body"].update(body_table)
                pane["footer"].update(
                    Panel(
                        Text.assemble(
                            ("Chain: ", "bright_magenta bold"),
                            blockchain_info.chain.ljust(9),
                            ("Blocks: ", "green bold"),
                            f"{blockchain_info.blocks}".ljust(8),
                            ("Size: ", "blue bold"),
                            f"{blockchain_info.size_on_disk}".ljust(10),
                            ("Time: ", "cyan bold"),
                            f"{blockchain_info.time}".rjust(10),
                        )
                    )
                )


__all__ = ["mine"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/nodekeys.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 05:31
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import List

### Third-party packages ###
from click import command
from docker import DockerClient, from_env
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich.progress import track

### Local modules ###
from src.schemas import LNDInfo


@command
def nodekeys() -> None:
    """Fetch nodekeys from active LND containers."""
    client: DockerClient = from_env()
    if client.ping():
        containers: List[Container] = reversed(client.containers.list())  # type: ignore[assignment]
        lnds: List[Container] = list(
            filter(lambda c: match(r"aesir-lnd|aesir-ping|aesir-pong", c.name), containers)
        )
        outputs: List[str] = []
        for container in track(lnds, "Fetch LND nodekeys:".ljust(42)):
            lnd_info: LNDInfo = TypeAdapter(LNDInfo).validate_json(
                container.exec_run(
                    """
                    lncli
                        --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                        --rpcserver=localhost:10001
                        --tlscertpath=/home/lnd/.lnd/tls.cert
                    getinfo
                    """
                ).output
            )
            outputs.append(f"<Nodekey: '{container.name}', '{lnd_info.identity_pubkey}'>")
        list(map(print, outputs))


__all__ = ["nodekeys"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    core.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 02:20
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from click import group

### Local modules ###
from src.commands import clean, deploy, flush, mine, nodekeys, ping_pong, setup


@group
def cli() -> None:
    """aesir"""


cli.add_command(clean, "clean")
cli.add_command(deploy, "deploy")
cli.add_command(flush, "flush")
cli.add_command(mine, "mine")
cli.add_command(nodekeys, "nodekeys")
cli.add_command(ping_pong, "ping-pong")
cli.add_command(setup, "setup")


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/commands/ping_pong.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-01 06:18
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import Dict, List

### Third-party packages ###
from click import argument, command
from docker import DockerClient, from_env
from docker.errors import NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter, ValidationError
from rich.progress import track

### Local modules ###
from src.schemas import LNDInfo, NewAddress, OpenChannel


@command
@argument("channel_size", default=16777215)
def ping_pong(channel_size: int) -> None:
    """For "duo" cluster, create channels between LND nodes."""
    client: DockerClient = from_env()
    if client.ping():
        bitcoind: Container
        try:
            bitcoind = client.containers.get("aesir-bitcoind")
        except NotFound:
            print('!! Unable to find "aesir-bitcoind" container.')
            return
        containers: List[Container] = list(reversed(client.containers.list()))
        paddles: List[Container] = list(
            filter(lambda c: match(r"aesir-ping|aesir-pong", c.name), containers)
        )
        mining_targets: Dict[str, str] = {}
        nodekeys: Dict[str, str] = {}

        ### Fetch nodekeys ###
        for container in track(paddles, "Fetch LND nodekeys:".ljust(42)):
            lnd_info: LNDInfo = TypeAdapter(LNDInfo).validate_json(
                container.exec_run(
                    """
                    lncli
                        --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                        --rpcserver=localhost:10001
                        --tlscertpath=/home/lnd/.lnd/tls.cert
                    getinfo
                    """
                ).output
            )
            nodekeys[container.name] = lnd_info.identity_pubkey
            new_address: NewAddress = TypeAdapter(NewAddress).validate_json(
                container.exec_run(
                    """
                    lncli
                        --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                        --rpcserver=localhost:10001
                        --tlscertpath=/home/lnd/.lnd/tls.cert
                    newaddress p2wkh
                    """
                ).output
            )
            mining_targets[container.name] = new_address.address

        ### Open channels ###
        outputs: List[str] = []
        for container in track(paddles, "Open channels:".ljust(42)):
            if container.name == "aesir-ping":
                try:
                    open_channel: OpenChannel = TypeAdapter(OpenChannel).validate_json(
                        container.exec_run(
                            """
                            lncli
                                --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                                --rpcserver=localhost:10001
                                --tlscertpath=/home/lnd/.lnd/tls.cert
                            openchannel %d
                                --node_key %s
                                --connect aesir-pong:9735
                            """
                            % (channel_size, nodekeys.get("aesir-pong", ""))
                        ).output
                    )
                    outputs.append(
                        f"<Channel 'aesir-ping --> aesir-pong' : { open_channel.funding_txid }>"
                    )
                    bitcoind.exec_run(
                        """
                        bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir generatetoaddress %d %s
                        """
                        % (6, mining_targets.get("aesir-ping", ""))
                    )
                except ValidationError:
                    outputs.append("!! Channel 'aesir-ping --> aesir-pong' already opened.")
            elif container.name == "aesir-pong":
                try:
                    open_channel: OpenChannel = TypeAdapter(OpenChannel).validate_json(  # type: ignore[no-redef]
                        container.exec_run(
                            """
                            lncli
                                --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                                --rpcserver=localhost:10001
                                --tlscertpath=/home/lnd/.lnd/tls.cert
                            openchannel %d
                                --node_key %s
                                --connect aesir-ping:9735
                            """
                            % (channel_size, nodekeys.get("aesir-ping", ""))
                        ).output
                    )
                    outputs.append(
                        f"<Channel 'aesir-pong --> aesir-ping' : { open_channel.funding_txid }>"
                    )
                    bitcoind.exec_run(
                        """
                        bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir generatetoaddress %d %s
                        """
                        % (6, mining_targets.get("aesir-pong", ""))
                    )
                except ValidationError:
                    outputs.append("!! Channel 'aesir-pong --> aesir-ping' already opened.")
        list(map(print, outputs))


__all__ = ["ping_pong"]



#!/usr/bin/env python3.8
# coding:utf-8
# Copyright (C) 2022-2023 All rights reserved.
# FILENAME:    ~~/src/schemas/peripheral_enum.py
# VERSION: 	   0.2.5
# CREATED: 	   2023-12-03 01:11
# AUTHOR: 	   Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Literal

PeripheralEnum = Literal["postgres", "redis"]

__all__ = ["PeripheralEnum"]
